close all;
clear all;

% Load the model from the obj file

filename = 'elephant';
% specify the number of eigenvalues/vectors to be calculated
nrOfEig = 50;

[vertices, faceVertexIndices] = importObj(sprintf('%s.obj',filename));
vCnt = length(vertices);
fCnt = length(faceVertexIndices);

% put the model into the TriRep Matlab class representation
trModel = TriRep(double(faceVertexIndices), vertices(:,1), vertices(:,2), vertices(:,3));

%%

% Build the needed neighbourhood information
edgeMidPt = zeros(vCnt, vCnt, 3); % Midpoint of the edge between v_i and v_j
neighV = cell(vCnt,1); % Neighbour vertices of v_i
neighE = cell(vCnt,1); % Neighbour edges of v_i

for v = 1:vCnt
   % Get the neighbourhood of simplices (triangles) of the actual vertex
   neighTri =  cell2mat(trModel.vertexAttachments(v));
   % Get the neighbouring vertices
   neighV{v,1} = unique(trModel.Triangulation(neighTri,:));
   neighV{v,1} = neighV{v}(neighV{v} ~= v); % Delete the actual vertex
   % Get the neighbouring edges
   neighE{v,1} = edgeAttachments(trModel, repmat(v, length(neighV{v}),1), neighV{v});
   
   % Calculate the midpoint of all edges (needed for the dual area)
   for vNIdx = 1:length(neighV{v})
       edgeMidPt(v, neighV{v}(vNIdx), :) = (trModel.X(v,:) + trModel.X(neighV{v}(vNIdx),:)) / 2;
   end
end

% Calculate the barycenter of all triangles (needed for the dual area)
faceCenter = zeros(fCnt,3);
for f = 1:fCnt
    vIdx = trModel.Triangulation(f,:);
    
    % Calculate the barycenter of the triangle face
    faceCenter(f,:) = sum(trModel.X(vIdx, :)) ./ 3;
end
%% plot the face centers and midpoints of the edges
% figure
% axis equal
% hold on
% trimesh(trModel);
% plot3(faceCenter(:,1), faceCenter(:,2), faceCenter(:,3), 'rx');
% plot3(edgeMidPt(:,:,1), edgeMidPt(:,:,2), edgeMidPt(:,:,3), 'bx');

%% Calculate the dual-Area and the cotan

cotMat = zeros(vCnt);
dualVol = zeros(vCnt, 1);

for vi = 1:length(vertices)
    % neighbouring vertex cnt
    neighVCnt = length(neighV{vi});

    %% Calculation of the Cotan
    
    % Go through all neighbouring vertices
    for vjIdx = 1:neighVCnt
       % Get the real vertex index of the neighbouring vertex
       vj = neighV{vi}(vjIdx);
       % Get the simplices bounded to an edge (should be at a max 2)
       comSimp = trModel.edgeAttachments(vi, vj);
       % Get the Vertices opposite of v_i and v_j
       oppV = unique(trModel.Triangulation(comSimp{:},:));
       oppV = oppV(oppV ~= vi & oppV ~= vj);

       % Calculate the cotan of both opposite angles
       if length(oppV) == 1
           viC = trModel.X(vi,:);
           vjC = trModel.X(vj,:);
           oppVC = trModel.X(oppV(1),:);

           beta(1) = vectorAngle(viC - oppVC, vjC - oppVC);

           cotMat(vi, vj) = cot(abs(beta(1)));
           %cotMat(vi, vj) = cot(beta(1));
       elseif length(oppV) == 2
           viC = trModel.X(vi,:);
           vjC = trModel.X(vj,:);
           oppVC1 = trModel.X(oppV(1),:);
           oppVC2 = trModel.X(oppV(2),:);

           beta(1) = vectorAngle(viC - oppVC1, vjC - oppVC1);
           beta(2) = vectorAngle(viC - oppVC2, vjC - oppVC2);

           cotMat(vi, vj) = cot(abs(beta(1))) + cot(abs(beta(2)));
           %cotMat(vi, vj) = cot(beta(1)) + cot(beta(2));
       else
           error('More than two opposite trinagles are found for one edge')
       end

    end
    
    %% Calculation of the dual Area
    
    % Go threw all neighbouring simplices in a specific order to calculate
    % the dual area
    
    % Add the face centers to the dual area to be calculated
    volPts = zeros(neighVCnt*2,3);
    
    % get the first neighbouring vertex to start at some point
    vj = neighV{vi}(1);
    
    
    neighTriIdx = cell2mat(trModel.edgeAttachments(vi, vj));
    neighTriIdx = neighTriIdx(1);
    
    volPtsIdx = 1;
    for i = 1:length(neighV{vi})
        if i > 1
           neighTriIdx = cell2mat(trModel.edgeAttachments(vi, vj));
           % Get the next triangle
           neighTriIdx = neighTriIdx(neighTriIdx ~= lastTriIdx);
        end
        
        % Get the vertices of the actual tri
        neighTriV = trModel.Triangulation(neighTriIdx,:);
        
        % Get the next vertex to go on
        lastTriIdx = neighTriIdx;
        lastVj = vj;
        vj = neighTriV(neighTriV ~= vi & neighTriV ~= vj);
        
        volPts(volPtsIdx, :) = edgeMidPt(vi, lastVj, :); 
        volPtsIdx = volPtsIdx+1;
        volPts(volPtsIdx, :) = faceCenter(neighTriIdx, :); 
        volPtsIdx = volPtsIdx+1;
%         volPts(volPtsIdx, :) = edgeMidPt(vi, vj, :); 
%         volPtsIdx = volPtsIdx+1;
    end
    
    volPtsMean = mean(volPts);
    for i = 1:length(volPts)
        if i < length(volPts)-1
            dualVol(vi) = dualVol(vi) + triArea(volPts(i,:)', volPts(i+1,:)', volPtsMean');
        else
            dualVol(vi) = dualVol(vi) + triArea(volPts(i,:)', volPts(1,:)', volPtsMean');
        end
    end
end

%% Calculate the Laplace-de Rham operator

delta = zeros(vCnt);
for vi = 1:vCnt
    for vj = 1:vCnt
        if vi ~= vj
            %delta(vi,vj) = - cotMat(vi,vj);
            delta(vi,vj) =  (-1) * cotMat(vi,vj)/sqrt(dualVol(vi)*dualVol(vj));
        end
    end
end

% Calculate the diagonal entries of the laplace-de rham operator
diagEnt = (-1)*sum(delta, 2);
delta = delta + diag(diagEnt);


%% Calculate the eigenvectors/-values
[eigVec eigVal] = eigs(delta, vCnt);
% Flip the eigenvalues/vectors such the the smallest is the first one
eigVal = flipud(diag(eigVal));
eigVec = fliplr(eigVec);

%% Write the file
eigVecFID = fopen(sprintf('%s_vectors.mhb',filename), 'w');
fwrite(eigVecFID, eigVec(:,1:nrOfEig), 'float32');
fclose(eigVecFID);

eigValFID = fopen(sprintf('%s_values.mhb',filename), 'w');
fwrite(eigValFID, eigVal(1:nrOfEig), 'float32');
fclose(eigValFID);
